# -*- coding: utf-8 -*-
#
# Decorator testing
# 
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
import unittest
from datetime import datetime
from gems import require, exception, keywords, cached



# tests
# -----
class TestRequire(unittest.TestCase):
    _a = None

    def requirement(self):
        self._a = 1
        return None

    @property
    @require('requirement')
    def property(self):
        return self._a

    def test_require(self):
        self.assertEqual(self._a, None)
        self.assertEqual(self.property, 1)
        return


class CacheExample(object):

    @cached.tag('reset')
    def time(self):
        return str(datetime.now().time())

    @cached.tag('reset')
    def ftime(self):
        return 'time: ' + str(datetime.now().time())


class CacheChildExample(CacheExample):

    @cached.tag('reset')
    def ctime(self):
        return str(datetime.now().time())


class TestCached(unittest.TestCase):

    def test_cached(self):
        # multiple properties
        foo = CacheExample()
        time = foo.time
        ftime = foo.ftime
        self.assertEqual(foo.time, time)
        self.assertEqual(foo.ftime, ftime)
        cached.invalidate(foo, 'reset')
        newtime = foo.time
        newftime = foo.ftime
        self.assertNotEqual(newtime, time)
        self.assertNotEqual(newftime, ftime)
        self.assertEqual(newtime, foo.time)
        self.assertEqual(newftime, foo.ftime)
        
        # multiple classes
        bar = CacheExample()
        btime = bar.time
        self.assertEqual(bar.time, btime)
        self.assertEqual(newtime, foo.time)
        cached.invalidate(bar, 'reset')
        finaltime = foo.time
        newbtime = bar.time
        self.assertNotEqual(newbtime, btime)
        self.assertEqual(finaltime, newtime)
        self.assertEqual(newbtime, bar.time)

        # object independence
        cached.invalidate(foo, 'reset')
        self.assertNotEqual(finaltime, foo.time)
        return

    def test_inheritance_cached(self):
        foo = CacheChildExample()
        time = foo.time
        ctime = foo.ctime
        self.assertEqual(foo.time, time)
        self.assertEqual(foo.ctime, ctime)
        cached.invalidate(foo, 'reset')
        newtime = foo.time
        newctime = foo.ctime
        self.assertNotEqual(newtime, time)
        self.assertNotEqual(newctime, ctime)
        self.assertEqual(newtime, foo.time)
        self.assertEqual(newctime, foo.ctime)
        return


class CustomException(Exception):
    pass


class TestException(unittest.TestCase):

    @exception(CustomException)
    def throw(self):
        raise AssertionError('This is an exception!')

    def test_exception(self):
        with self.assertRaises(CustomException):
            self.throw()
        return


class TestKeywords(unittest.TestCase):

    @keywords
    def function(self, *args, **kwargs):
        return args, kwargs

    def test_keywords(self):
        args, kwargs = self.function(one=1, two=2)
        self.assertEqual(len(args), 0)
        self.assertEqual(kwargs, {'one': 1, 'two': 2})

        args, kwargs = self.function(1, two=2)
        self.assertEqual(len(args), 1)
        self.assertEqual(kwargs, {'two': 2})

        args, kwargs = self.function({'one': 1}, two=2)
        self.assertEqual(len(args), 0)
        self.assertEqual(kwargs, {'one': 1, 'two': 2})

        args, kwargs = self.function({'one': 1}, two=2)
        self.assertEqual(len(args), 0)
        self.assertEqual(kwargs, {'one': 1, 'two': 2})
        return



# -*- coding: utf-8 -*-
#
# Useful metaclasses.
#
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
# none yet ...


# metaclasses
# -----------
class DocRequire(type):
    """
    Metaclass forcing requirement of docstrings on all public
    class methods.

    Example:
        >>> # class without docstrings will throw
        >>> # a TypeError
        >>> class A(object):
        >>>     def method(self):
        >>>         return 1
        >>>
        TypeError: method must have a docstring

        >>> # class with docstrings is fine
        >>> class A(object):
        >>>     ''' A class with docstrings '''
        >>>     def method(self):
        >>>         ''' A method  '''
        >>>         return 1
        >>>
    """

    def __init__(self, name, bases, attrs):
        for key, value in attrs.items():
            if key.startswith("__"):
                continue
            if not hasattr(value, "__call__"):
                continue
            if not getattr(value, '__doc__'):
                raise AssertionError("%s must have a docstring" % key)
        type.__init__(self, name, bases, attrs)



# -*- coding: utf-8 -*-
#
# Useful decorators.
#
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
from functools import wraps
import inspect
from future.utils import raise_with_traceback
from .datatypes import composite


# class-related
# -------------
def require(method):
    """
    Decorator for managing chained dependencies of different class
    properties. The @require decorator allows developers to specify
    that a function call must be operated on before another property
    or function call is accessed, so that data and processing for an
    entire class can be evaluated in a lazy way (i.e. not all upon
    instantiation).

    Examples:

        >>> class Foo(Bar):
        >>>
        >>>    def a(self):
        >>>        print 'a!'
        >>>        return 1
        >>>
        >>>    @require('a')
        >>>    @property
        >>>    def b(self):
        >>>        print 'b!'
        >>>        return self.a + 1
        >>>
        >>> foo = Foo()
        >>> print foo.b
        >>>
        'a!'
        'b!'
        2
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # throw exception if input class doesn't have requirement
            if not hasattr(args[0], method):
                raise AssertionError('{} class has no method {}()'.format(args[0].__class__.__name__, method))

            # create property to record that method has been called
            callmethod = method + '_called'
            if not hasattr(args[0], callmethod):
                setattr(args[0], callmethod, False)

            # call the method if it hasn't yet been called
            if not getattr(args[0], callmethod):
                getattr(args[0], method)()
                setattr(args[0], callmethod, True)
            return func(*args, **kwargs)
        return wrapper
    return decorator


class cached(object):
    """
    Special implementation of cached_property by pydanny to allow for
    tagging and invalidation of caches.

    Examples:
        >>> class Foo(object):
        >>> @cached.tag('stateful')
        >>> def time(self)
        >>>     return str(datetime.now().time())
        >>>
        >>> obj = Foo()
        >>> obj.time
        '11:24:33.613401'
        >>> time.sleep(5)
        >>> obj.time
        '11:24:33.613401'
        >>> cached.invalidate(obj, 'stateful')
        >>> obj.time
        '11:26:35.197487'
    """

    def __init__(self, func, *tags):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func
        self.tags = frozenset(tags)
        return

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value

    @classmethod
    def tag(cls, *tags):
        return lambda f: cls(f, *tags)

    @staticmethod
    def invalidate(obj, tag):
        for base in obj.__class__.mro():
            for key, value in base.__dict__.items():
                if isinstance(value, cached) and tag in value.tags:
                    obj.__dict__.pop(key, None)
        return


# exception-related
# -----------------
def exception(exception):
    """
    Wrap function/method with specific exception if any
    exception occurs during function execution.
 
    Args:
        exception (Exception): Exception to re-cast error as.

    Examples:
        >>> from gems import exception
        >>>
        >>> class MyCustomException(Exception):
        >>>     pass
        >>>
        >>> @exception(MyCustomException)
        >>> def func():
        >>>     return 1 / 0
        >>>
        >>> func()
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "gems/decorators.py", line 96, in wrapper
            return func(*args, **kwargs)
          File "<stdin>", line 3, in func
        __main__.MyCustomException: integer division or modulo by zero
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as exe:
                raise raise_with_traceback(exception(exe))
        return wrapper
    return decorator


# function-related
# ----------------
def keywords(func):
    """
    Accumulate all dictionary and named arguments as
    keyword argument dictionary. This is generally useful for
    functions that try to automatically resolve inputs.

    Examples:
        >>> @keywords
        >>> def test(*args, **kwargs):
        >>>     return kwargs
        >>>
        >>> print test({'one': 1}, two=2)
        {'one': 1, 'two': 2}
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        idx = 0 if inspect.ismethod(func) else 1
        if len(args) > idx:
            if isinstance(args[idx], (dict, composite)):
                for key in args[idx]:
                    kwargs[key] = args[idx][key]
                args = args[:idx]
        return func(*args, **kwargs)
    return decorator



# -*- coding: utf-8 -*-
#
# Decorators for terminal-based wait animations
#
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
import os
import re
import json
import yaml
import glob

from .utils import depricated_name, deprecated


# data management
# -----------------
class composite(object):
    """
    Data structure for traversing object relationships via attributes
    instead of keys and indices.

    Args:
        data (tuple, list, dict): Data to build composite datastructure from.

    Example:
        >>> data = composite({
        >>>     'one': 1,
        >>>     'two': [1, 2, 3],
        >>>     'three': ['one', 2, {'three': 'four'}],
        >>>     'four': {'five': [6, 7, 8], 'nine': 10, 'eleven': 'twelve'}
        >>> })
        >>> data.four.five[1] == 6
        True
        >>> data.two[0] == 1
        True
    """

    def __init__(self, data):
        self._list = []
        self._dict = {}
        self.meta_type = None

        if hasattr(data, 'read'):
            data = json.load(data)

        if isinstance(data, (list, tuple)):
            for dat in data:
                if not isinstance(dat, (list, tuple, dict)):
                    self._list.append(dat)
                else:
                    self._list.append(composite(dat))
            self.meta_type = 'list'

        elif isinstance(data, dict):
            for key in data:
                if not isinstance(data[key], (list, tuple, dict)):
                    self._dict[key] = data[key]
                else:
                    self._dict[key] = composite(data[key])
            self.meta_type = 'dict'

        elif isinstance(data, composite):
            self.meta_type = data.meta_type
            if data.meta_type == 'dict':
                self._dict = data._dict
            elif data.meta_type == 'list':
                self._list = data._list

        else:
            raise TypeError('unsupported type for creating composite: {}'.format(type(data)))
        return

    @classmethod
    def load(cls, fh):
        """
        Load json or yaml data from file handle.

        Args:
            fh (file): File handle to load from.

        Examlple:
            >>> with open('data.json', 'r') as json:
            >>>    jsdata = composite.load(json)
            >>>
            >>> with open('data.yml', 'r') as yml:
            >>>    ymldata = composite.load(yml)
        """
        dat = fh.read()
        try:
            ret = cls.from_json(dat)
        except:
            ret = cls.from_yaml(dat)
        return ret

    @classmethod
    def from_json(cls, fh):
        """
        Load json from file handle.

        Args:
            fh (file): File handle to load from.

        Examlple:
            >>> with open('data.json', 'r') as json:
            >>>    data = composite.load(json)
        """
        if isinstance(fh, str):
            return cls(json.loads(fh))
        else:
            return cls(json.load(fh))

    @classmethod
    def from_yaml(cls, fh):
        """
        Load yaml from file handle.

        Args:
            fh (file): File handle to load from.

        Examlple:
            >>> with open('data.yml', 'r') as json:
            >>>    data = composite.load(json)
        """
        return cls(yaml.load(fh))

    @classmethod
    def from_string(cls, string):
        """
        Load data from string.

        Args:
            string (str): String to load from.

        Examlple:
            >>> with open('data.json', 'r') as json:
            >>>     jdat = json.read()
            >>> data = composite.from_string(jdat)
        """
        return cls(eval(string))

    def __len__(self):
        return max(len(self._list), len(self._dict))

    def __str__(self):
        return str(self.json())

    def __repr__(self):
        return str(self)

    def __iter__(self):
        if self.meta_type == 'list':
            for entry in self._list:
                yield entry

        elif self.meta_type == 'dict':
            for entry in self._dict:
                yield entry
        return

    def __getattr__(self, name):
        if name in self._dict:
            return self._dict[name]
        else:
            raise AttributeError('\'composite\' object has no attribute {}'.format(name))
        return

    def __getitem__(self, item):
        if self.meta_type == 'list':
            return self._list[item]
        elif self.meta_type == 'dict':
            return self._dict[item]
        else:
            raise KeyError(str(item))
        return

    def __setattr__(self, name, value):
        if name == '_list' or name == '_dict' or name == 'meta_type':
            super(composite, self).__setattr__(name, value)
        else:
            self._dict[name] = value
        return

    def __setitem__(self, idx, value):
        if self.meta_type == 'list':
            self._list[idx] = value
        elif self.meta_type == 'dict':
            self._dict[idx] = value
        return

    def __add__(self, other):
        # TODO: Think about doing a recursive addition of all the
        #       properties -- for strings, concat, for ints, add, etc.
        #       Get feedback about this, and see if it intuitively makes
        #       sense. Since we have set-based operators now, it makes
        #       sense.
        if self.meta_type == 'list':
            if isinstance(other, (composite, dict, list, tuple)):
                other = composite(other)
                if len(self) == 0:
                    return other
                elif len(other) == 0:
                    return self
                elif other.meta_type == 'list':
                    return composite(self._list + other._list)
                elif other.meta_type == 'dict':
                    return composite([self._list, other._dict])
            else:
                return composite(self._list + [other])
        elif self.meta_type == 'dict':
            if isinstance(other, (composite, dict, list, tuple)):
                other = composite(other)
                if len(self) == 0:
                    return other
                elif len(other) == 0:
                    return self
                elif other.meta_type == 'list':
                    return composite([self._dict, other._list])
                elif other.meta_type == 'dict':
                    return self.union(other, recursive=True, overwrite=True)
            else:
                return composite([self._dict, other])
        return

    def __contains__(self, item):
        if self.meta_type == 'list':
            return item in self._list
        elif self.meta_type == 'dict':
            return item in self._dict

    def __eq__(self, other):
        if isinstance(other, composite):
            return self.json() == other.json()
        else:
            return self.json() == other

    def __ne__(self, other):
        return not (self == other)

    def intersection(self, other, recursive=True):
        """
        Recursively compute intersection of data. For dictionaries, items
        for specific keys will be reduced to unique items. For lists, items
        will be reduced to unique items. This method is meant to be analogous
        to set.intersection for composite objects.

        Args:
            other (composite): Other composite object to intersect with.
            recursive (bool): Whether or not to perform the operation recursively,
                for all nested composite objects.
        """
        if not isinstance(other, composite):
            raise AssertionError('Cannot intersect composite and {} types'.format(type(other)))
        
        if self.meta_type != other.meta_type:
            return composite({})

        if self.meta_type == 'list':
            keep = []
            for item in self._list:
                if item in other._list:
                    if recursive and isinstance(item, composite):
                        keep.extend(item.intersection(other.index(item), recursive=True))
                    else:
                        keep.append(item)
            return composite(keep)
        elif self.meta_type == 'dict':
            keep = {}
            for key in self._dict:
                item = self._dict[key]
                if key in other._dict:
                    if recursive and \
                       isinstance(item, composite) and \
                       isinstance(other.get(key), composite):
                       keep[key] = item.intersection(other.get(key), recursive=True)
                    elif item == other[key]:
                        keep[key] = item
            return composite(keep)
        return

    def difference(self, other, recursive=True):
        """
        Recursively compute difference of data. For dictionaries, items
        for specific keys will be reduced to differences. For lists, items
        will be reduced to differences. This method is meant to be analogous
        to set.difference for composite objects.

        Args:
            other (composite): Other composite object to difference with.
            recursive (bool): Whether or not to perform the operation recursively,
                for all nested composite objects.
        """
        if not isinstance(other, composite):
            raise AssertionError('Cannot difference composite and {} types'.format(type(other)))
        
        if self.meta_type != other.meta_type:
            return self

        if self.meta_type == 'list':
            keep = []
            for item in self._list:
                if item not in other._list:
                    if recursive and isinstance(item, composite):
                        keep.extend(item.difference(other.index(item), recursive=True))
                    else:
                        keep.append(item)
            return composite(keep)
        elif self.meta_type == 'dict':
            keep = {}
            for key in self._dict:
                item = self._dict[key]
                if key in other._dict:
                    if recursive and \
                       isinstance(item, composite) and \
                       isinstance(other.get(key), composite):
                       keep[key] = item.difference(other.get(key), recursive=True)
                    elif item != other[key]:
                        keep[key] = item
                else:
                    keep[key] = item
            return composite(keep)
        return

    def union(self, other, recursive=True, overwrite=False):
        """
        Recursively compute union of data. For dictionaries, items
        for specific keys will be combined into a list, depending on the
        status of the overwrite= parameter. For lists, items will be appended
        and reduced to unique items. This method is meant to be analogous
        to set.union for composite objects.

        Args:
            other (composite): Other composite object to union with.
            recursive (bool): Whether or not to perform the operation recursively,
                for all nested composite objects.
            overwrite (bool): Whether or not to overwrite entries with the same
                key in a nested dictionary. 
        """
        if not isinstance(other, composite):
            raise AssertionError('Cannot union composite and {} types'.format(type(other)))
        
        if self.meta_type != other.meta_type:
            return composite([self, other])

        if self.meta_type == 'list':
            keep = []
            for item in self._list:
                keep.append(item)
            for item in other._list:
                if item not in self._list:
                    keep.append(item)
            return composite(keep)
        elif self.meta_type == 'dict':
            keep = {}
            for key in list(set(list(self._dict.keys()) + list(other._dict.keys()))):
                left = self._dict.get(key)
                right = other._dict.get(key)
                if recursive and \
                   isinstance(left, composite) and \
                   isinstance(right, composite):
                    keep[key] = left.union(right, recursive=recursive, overwrite=overwrite)
                elif left == right:
                    keep[key] = left
                elif left is None:
                    keep[key] = right
                elif right is None:
                    keep[key] = left
                elif overwrite:
                    keep[key] = right
                else:
                    keep[key] = composite([left, right])
            return composite(keep)
        return

    def index(self, item):
        """
        Return index containing value.
        """
        return self._list.index(item)

    def get(self, *args, **kwargs):
        """
        Return item or None, depending on if item exists. This is
        meant to be similar to dict.get() for safe access of a property.
        """
        return self._dict.get(*args, **kwargs)

    def update(self, other):
        """
        Update internal dictionary object. This is meant to be an
        analog for dict.update().
        """
        if self.meta_type == 'list':
            raise AssertionError('Cannot update object of `list` base type!')
        elif self.meta_type == 'dict':
            self._dict = dict(self + composite(other))
            return

    def keys(self):
        """
        Return keys for object, if they are available.
        """
        if self.meta_type == 'list':
            return None
        elif self.meta_type == 'dict':
            return self._dict.keys()

    def items(self):
        """
        Return keys for object, if they are available.
        """
        if self.meta_type == 'list':
            return self._list
        elif self.meta_type == 'dict':
            return self._dict.items()

    def values(self):
        """
        Return keys for object, if they are available.
        """
        if self.meta_type == 'list':
            return self._list
        elif self.meta_type == 'dict':
            return self._dict.values()

    def append(self, item):
        """
        Append to object, if object is list.
        """
        if self.meta_type == 'dict':
            raise AssertionError('Cannot append to object of `dict` base type!')
        if self.meta_type == 'list':
            self._list.append(item)
        return

    def extend(self, item):
        """
        Extend list from object, if object is list.
        """
        if self.meta_type == 'dict':
            raise AssertionError('Cannot extend to object of `dict` base type!')
        if self.meta_type == 'list':
            self._list.extend(item)
        return

    def json(self):
        """
        Return JSON representation of object.
        """
        if self.meta_type == 'list':
            ret = []
            for dat in self._list:
                if not isinstance(dat, composite):
                    ret.append(dat)
                else:
                    ret.append(dat.json())
            return ret

        elif self.meta_type == 'dict':
            ret = {}
            for key in self._dict:
                if not isinstance(self._dict[key], composite):
                    ret[key] = self._dict[key]
                else:
                    ret[key] = self._dict[key].json()
            return ret

    def write_json(self, fh, pretty=True):
        """
        Write composite object to file handle in JSON format.

        Args:
            fh (file): File handle to write to.
            pretty (bool): Sort keys and indent in output.
        """
        sjson = json.JSONEncoder().encode(self.json())
        if pretty:
            json.dump(json.loads(sjson), fh, sort_keys=True, indent=4)
        else:
            json.dump(json.loads(sjson), fh)
        return

    def write_yaml(self, fh):
        """
        Write composite object to file handle in YAML format.

        Args:
            fh (file): File handle to write to.
        """
        yaml.dump(self.json(), fh)
        return

    def write(self, fh, pretty=True):
        """
        API niceness defaulting to composite.write_json().
        """
        return self.write_json(fh, pretty=pretty)


# data management
# -----------------
class filetree(object):
    """
    Data structure for traversing directory structure and creating object for
    accessing relative file paths.

    .. NOTE: The filetree is completely walked through when this object is
        instantiated, so expect object creation for large trees to be
        relatively expensive.

    Args:
        directory (str): Directory to build filetree from.
        ignore (str): Regular expression with items to ignore. If you wish
            to recurse through all directories (including hidden directories),
            set ignore=None. By default, this is set to "^[._]" (i.e. any files
            beginning with "." or "_").

    Example:
        >>> data = filetree('mydir')
        >>> print data
        mydir/
             one/
                two.txt
                three.json
            two/
                three/
                      four.txt
                five six/
                         seven.txt
                eight.config
        >>> print data.one['two.txt']
        /full/path/to/mydir/one/two.txt
        >>> print data.two.three['four.txt']
        /full/path/to/mydir/two/three/four.txt
        >>> print data.two['five six']['eight.config']
        /full/path/to/mydir/two/five six/eight.config
    """

    def __init__(self, directory, ignore=r"^[._]", regex=r".*"):
        self.ignore = ignore
        self.regex = regex
        self._data = {}
        self.root = os.path.realpath(directory)
        if os.path.isdir(directory):
            for item in os.listdir(self.root):
                if ignore is not None:
                    if re.search(ignore, item):
                        continue
                fullpath = os.path.realpath(os.path.join(self.root, item))
                if os.path.isdir(fullpath):
                    self._data[os.path.basename(fullpath)] = filetree(fullpath, ignore=ignore, regex=regex)
                else:
                    if re.search(regex, fullpath):
                        self._data[os.path.basename(fullpath)] = fullpath
        else:
            for fi in glob.glob(directory):
                fullpath = os.path.realpath(fi)
                self._data[os.path.basename(fullpath)] = fullpath
        self._filelist = []
        return

    def __len__(self):
        return len(self._data)

    def __str__(self):
        """
        .. NOTE:: This needs to be completed -- print filetree
        """
        def jstr(data, tab=''):
            res = ''
            for item in data:
                if isinstance(data[item], basestring):
                    res += tab + item + '\n'
                else:
                    res += tab + item + '/\n'
                    res += jstr(data[item], tab=tab+'\t')
            return res
        return jstr(self.json())

    def __iter__(self):
        for item in self._data:
            yield self._data[item]

    def __getattr__(self, name):
        if name not in self._data:
            raise AttributeError('filetree object has no attribute {}!'.format(name))
        return self._data[name]

    def __getitem__(self, item):
        if item not in self._data:
            raise KeyError('filetree object has no attribute {}!'.format(item))
        return self._data[item]

    def __contains__(self, item):
        if item in self._data:
            return True
        item = item.replace(self.root, '')
        item = re.sub(r"^[/\\]", "", item)
        subpaths = item.split("/")
        idx, cdata = 0, self
        try:
            while idx < len(subpaths):
                cdata = cdata[subpaths[idx]]
                idx += 1
        except KeyError:
            return False
        return True

    def __eq__(self, other):
        sf = sorted(self.filelist())
        of = sorted(other.filelist())
        if len(sf) != len(of):
            return False
        for i in range(0, len(sf)):
            if sf[i] != of[i]:
                return False
        return True

    def get(self, item):
        """
        Safe way to get items, similar to __dict__.get().

        Args:
            item (str): Item to get in file tree.
        """
        return self._data.get(item)

    def json(self):
        """
        Return JSON representation of object.
        """
        data = {}
        for item in self._data:
            if isinstance(self._data[item], filetree):
                data[item] = self._data[item].json()
            else:
                data[item] = self._data[item]
        return data

    @depricated_name('filelist()')
    def files(self):
        return self.filelist()

    def filelist(self):
        """
        Return list of files in filetree.
        """
        if len(self._filelist) == 0:
            for item in self._data:
                if isinstance(self._data[item], filetree):
                    self._filelist.extend(self._data[item].filelist())
                else:
                    self._filelist.append(self._data[item])
        return self._filelist

    def prune(self, regex=r".*"):
        """
        Prune leaves of filetree according to specified
        regular expression.

        Args:
            regex (str): Regular expression to use in pruning tree.
        """
        return filetree(self.root, ignore=self.ignore, regex=regex)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# testing for animation
# 
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
import os
import uuid
import unittest
from gems import composite, filetree


# config
# ______
__resources__ = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources')


# tests
# -----
class TestComposite(unittest.TestCase):
    _dict = {
        'one': 1,
        'two': [1, 2, 3],
        'three': ['one', 2, {'three': 'four'}],
        'four': {'five': [6, 7, 8], 'nine': 10, 'eleven': 'twelve'}
    }
    _list = [
        1, {'two': 3, 'four': 'five', 'six': [7, 8, 9, 'ten']},
        [11, 12, {'thirteen': 'fourteen', 'fifteen': 16, 'seventeen': [18]}]
    ]
    _c1 = composite({
        'one': 1,
        'two': [1, 2],
        'three': {'four': 5, 'five': 7},
        'eight': 8
    })
    _c2 = composite({
        'one': 1,
        'two': [1, 2, 3],
        'three': {'four': 5, 'six': 7},
        'eight': 9,
        'nine': 10
    })

    def test_load(self):
        data = composite(self._dict)
        self.assertEqual(sorted(data.keys()), sorted(self._dict.keys()))
        self.assertEqual(len(data), 4)
        data = composite.from_string(str(self._dict))
        self.assertEqual(sorted(data.keys()), sorted(self._dict.keys()))
        self.assertEqual(len(data), 4)
        with open(os.path.join(__resources__, 'dict.json'), 'r') as fi:
            data = composite.load(fi)
        self.assertEqual(sorted(data.keys()), sorted(self._dict.keys()))
        self.assertEqual(len(data), 4)

        data = composite(self._list)
        self.assertEqual(data.keys(), None)
        self.assertEqual(len(data), 3)
        data = composite.from_string(str(self._list))
        self.assertEqual(data.keys(), None)
        self.assertEqual(len(data), 3)
        with open(os.path.join(__resources__, 'list.json'), 'r') as fi:
            data = composite.load(fi)
        self.assertEqual(data.keys(), None)
        self.assertEqual(len(data), 3)
        return

    def test_write(self):
        data = composite(self._dict)
        fname = '.datatypes-test-' + str(uuid.uuid1()) + '.json'
        with open(fname, 'w') as of:
            data.write(of, pretty=True)
        with open(fname, 'r') as fi:
            data2 = composite.load(fi)
        self.assertEqual(data, data2)
        os.remove(fname)
        return

    def test_properties(self):
        data = composite(self._dict)
        self.assertEqual(len(data.items()), len(self._dict.items()))
        self.assertEqual(
            list(map(lambda x: x[1], data.items())),
            list(data.values())
        )
        self.assertEqual(data.one, 1)
        self.assertEqual(data['one'], 1)
        self.assertEqual(data.two[1], 2)
        self.assertEqual(data.three[2].three, 'four')
        self.assertEqual(data.four.five[2], 8)
        self.assertEqual(data.four.five[1:3], [7, 8])
        data.one = 2
        self.assertEqual(data.one, 2)
        self.assertEqual(data._dict['one'], 2)
        data['one'] = 3
        self.assertEqual(data.one, 3)
        data.five = 6
        self.assertEqual(data.five, 6)
        self.assertEqual(data._dict['five'], 6)
        data['five'] = 7
        self.assertEqual(data.five, 7)
        self.assertEqual(len(data.items()), len(self._dict.items()) + 1)

        data = composite(self._list)
        self.assertEqual(data[0], 1)
        self.assertEqual(data[1].six[3], 'ten')
        self.assertEqual(data[2][1], 12)
        self.assertEqual(data[2][2].seventeen[0], 18)
        self.assertEqual(data[1].six[0:2], [7, 8])
        data[0] = 2
        self.assertEqual(data[0], 2)
        self.assertEqual(len(data.items()), len(self._list))
        self.assertEqual(list(data.items()), list(data.values()))
        return

    def test_dict_methods(self):
        data = composite(self._dict)

        # get
        self.assertEqual(data.get('one'), 1)
        self.assertEqual(data.get('missing'), None)
        self.assertEqual(data.get('missing', 1), 1)

        # update
        data.update({
            'one': 2,
            'two': 'three',
            'five': {'six': 6}
        })
        self.assertEqual(data.one, 2)
        self.assertEqual(data.five.six, 6)
        self.assertEqual(data.two, 'three')
        return

    def test_iteration(self):
        data = composite(self._dict)
        self.assertEqual(sorted([i for i in data]), sorted(['one', 'two', 'three', 'four']))
        self.assertEqual([i for i in data.two], [1, 2, 3])

        data = composite(self._list)
        self.assertEqual([i for i in data][0], 1)
        self.assertEqual(sorted([i for i in data[2][2]]), sorted(['thirteen', 'fifteen', 'seventeen']))
        return

    def test_add(self):
        data = composite(self._dict)
        self.assertTrue(isinstance(data + {'five': 6}, composite))
        self.assertTrue(isinstance(data + 6, composite))
        self.assertTrue(isinstance(data + [1, 2], composite))
        self.assertTrue(isinstance(data + composite([1, 2]), composite))
        obj = data + {'five': 6}
        self.assertEqual(obj.one, 1)
        self.assertEqual(obj.five, 6)
        obj = data + {'one': 2}
        self.assertEqual(obj.one, 2)
        obj += {'one': 2}
        self.assertEqual(obj.one, 2)
        obj = data + 6
        self.assertEqual(obj[0].one, 1)
        self.assertEqual(obj[1], 6)
        obj = data + [1, 2]
        self.assertEqual(obj[0].one, 1)
        self.assertEqual(obj[1][1], 2)

        data = composite(self._list)
        self.assertTrue(isinstance(data + {'five': 6}, composite))
        self.assertTrue(isinstance(data + 6, composite))
        self.assertTrue(isinstance(data + [1, 2], composite))
        self.assertTrue(isinstance(data + composite([1, 2]), composite))
        obj = data + {'five': 6}
        self.assertEqual(obj[0][0], 1)
        self.assertEqual(obj[1].five, 6)
        obj = data + 6
        self.assertEqual(obj[-1], 6)
        self.assertEqual(obj[1].two, 3)
        obj = data + [1, 2]
        self.assertEqual(obj[-2:], [1, 2])
        self.assertEqual(obj[1].six[1], 8)

        obj = composite({}) + data
        self.assertEqual(obj[1].two, 3)

        obj = data + composite({})
        self.assertEqual(obj[1].two, 3)
        return

    def test_eq(self):
        data = composite(self._dict)
        self.assertEqual(data, self._dict)
        self.assertEqual(data, composite(self._dict))
        self.assertNotEqual(data, self._list)
        self.assertNotEqual(data, composite(self._list))
        return

    def test_exceptions(self):
        data = composite(self._dict)
        with self.assertRaises(KeyError):
            data['notakey']
        with self.assertRaises(AttributeError):
            data.notakey
        return

    def test_intersection(self):
        result = composite({
            'one': 1,
            'two': [1, 2],
            'three': {'four': 5},
        })
        self.assertEqual(self._c1.intersection(self._c2), result)
        empty = self._c1.intersection(composite({'five': 7}))
        self.assertTrue(isinstance(empty, composite))
        self.assertEqual(len(empty), 0)
        empty = composite({}).intersection(composite({}))
        self.assertTrue(isinstance(empty, composite))
        self.assertEqual(len(empty), 0)
        return

    def test_difference(self):
        result = composite({
            'two': [3],
            'three': {'six': 7},
            'eight': 9,
            'nine': 10
        })
        self.assertEqual(self._c2.difference(self._c1), result)
        return

    def test_union(self):
        result = composite({
            'one': 1,
            'two': [1, 2, 3],
            'three': {'four': 5, 'five': 7, 'six': 7},
            'eight': [8, 9],
            'nine': 10
        })
        self.assertEqual(self._c1.union(self._c2), result)
        return

    def test_filetypes(self):
        with open(os.path.join(__resources__, 'list.json'), 'r') as fi:
            js = composite.from_json(fi)
        with open(os.path.join(__resources__, 'list.yml'), 'r') as fi:
            yml = composite.from_yaml(fi)
        self.assertEqual(js, yml)

        with open(os.path.join(__resources__, 'dict.json'), 'r') as fi:
            js = composite.from_json(fi)
        with open(os.path.join(__resources__, 'dict.yml'), 'r') as fi:
            yml = composite.from_yaml(fi)
        self.assertEqual(js, yml)
        return

    def test_implicit_load(self):
        with open(os.path.join(__resources__, 'list.json'), 'r') as fi:
            js = composite.load(fi)
        with open(os.path.join(__resources__, 'list.yml'), 'r') as fi:
            yml = composite.load(fi)
        self.assertEqual(js, yml)
        return


class TestFiletree(unittest.TestCase):
    _dir = os.path.realpath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '..'
        )
    )

    def test_properties(self):
        data = filetree(self._dir)
        self.assertTrue(os.path.exists(data.tests['test_datatypes.py']))
        self.assertTrue('tests/test_datatypes.py' in data.tests['test_datatypes.py'])
        self.assertTrue(isinstance(data['tests'].resources, filetree))
        self.assertTrue(os.path.exists(data.tests.resources['dict.json']))
        self.assertTrue('tests/resources/dict.json' in data.tests.resources['dict.json'])
        
        data = filetree(os.path.join(__resources__, '*.json'))
        self.assertEqual(len(data), 2)
        return

    def test_operators(self):
        data = filetree(self._dir)
        self.assertTrue('tests/resources' in data)
        self.assertTrue('tests/resources/dict.json' in data)
        self.assertFalse('tests/resources/di.json' in data)
        self.assertFalse('fakefile' in data)
        return

    def test_get(self):
        data = filetree(self._dir)
        self.assertTrue('tests/test_datatypes.py' in data.tests.get('test_datatypes.py'))
        self.assertEqual(data.tests.get('test_datatypes.pypy'), None)
        return

    def test_json(self):
        data = filetree(self._dir)
        json = data.json()
        self.assertEqual(sorted(json.keys()), sorted([name for name in data._data]))
        self.assertEqual(json['tests']['resources']['dict.json'], data.tests.resources['dict.json'])
        return

    def test_match(self):
        data = filetree(self._dir, regex=r".*.yml$")
        self.assertEqual(len(data.filelist()), 2)
        self.assertTrue('dict.yml' in sorted(data.filelist())[0])
        self.assertTrue('list.yml' in sorted(data.filelist())[1])
        return

    def test_prune(self):
        raw = filetree(self._dir)
        data = filetree(self._dir, regex=r".*.json$")
        pruned = data.prune(r".*.json$")
        self.assertEqual(data, pruned)
        return

    def test_exceptions(self):
        raw = filetree(self._dir)
        with self.assertRaises(KeyError):
            raw['notafile']
        with self.assertRaises(AttributeError):
            raw.notafile
        return




# -*- coding: utf-8 -*-
#
# Decorators for managing repository
#
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
from functools import wraps
import warnings


# decorators
# ----------
def depricated_name(newmethod):
    """
    Decorator for warning user of depricated functions before use.

    Args:
        newmethod (str): Name of method to use instead.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning) 
            warnings.warn(
                "Function {} is depricated, please use {} instead.".format(func.__name__, newmethod),
                category=DeprecationWarning, stacklevel=2
            )
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def deprecated(func):
    """
    Decorator for warning user of depricated functions before use.
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        # warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            "Function {} is depricated. Consult the documentation for a better way of performing this task.".format(func.__name__),
            category=DeprecationWarning, stacklevel=2
        )
        # warnings.simplefilter('default', DeprecationWarning)
        return func(*args, **kwargs)
    return decorator



# -*- coding: utf-8 -*-

__author__ = 'Blake Printy'
__email__ = 'bprinty@gmail.com'
__version__ = '0.3.3'


from .datatypes import composite
from .datatypes import filetree

from .decorators import require
from .decorators import exception
from .decorators import keywords
from .decorators import cached

from .metaclasses import DocRequire



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# package setup
# 
# @author <bprinty@gmail.com>
# ------------------------------------------------


# config
# ------
import os
import re
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open(os.path.join('gems', '__init__.py'), 'r') as fi:
    __version__ = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fi.read()).group(1)


with open('requirements.txt', 'r') as reqs:
    requirements = map(lambda x: x.rstrip(), reqs.readlines())


test_requirements = [
    'pytest',
    'pytest-runner'
]


# files
# -----
with open('README.rst') as readme_file:
    readme = readme_file.read()


# exec
# ----
setup(
    name='gems',
    version=__version__,
    description="Python utilities for data manipulation and management.",
    long_description=readme,
    author="Blake Printy",
    author_email='bprinty@gmail.com',
    url='https://github.com/bprinty/gems',
    packages=['gems'],
    package_dir={'gems': 'gems'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache-2.0",
    zip_safe=False,
    keywords=['gems', 'data', 'structures', 'types', 'filesystem', 'management'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apple Public Source License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    tests_require=test_requirements,
    setup_requires=test_requirements
)



# -*- coding: utf-8 -*-
#
# Datatype testing
# 
# @author <bprinty@gmail.com>
# ------------------------------------------------


# imports
# -------
import unittest
from gems import DocRequire
import six


# tests
# -----
class TestDocRequire(unittest.TestCase):

    def test_docrequire(self):
        # raises AssertionError when evaled
        with self.assertRaises(AssertionError):
            
            @six.add_metaclass(DocRequire)
            class A(object):    
                def method(self):
                    return
        
        # shouldn't raise an AssertionError
        with self.assertRaises(TypeError):
            
            @six.add_metaclass(DocRequire)
            class A(object):
                def method(self):
                    """ okay """
                    return
            
            raise TypeError
        return
